import hmac
from typing import cast

if hasattr(hmac, "compare_digest"):
    # Prefer the stdlib implementation, when available.
    def secure_compare(val1, val2):
        return hmac.compare_digest(val1, val2)

else:

    def secure_compare(val1, val2):
        """
        Returns True if the two strings are equal, False otherwise.
        The time taken is independent of the number of characters that match.
        For the sake of simplicity, this function executes in constant time
        only when the two strings have the same length. It short-circuits when
        they have different lengths.
        """
        if len(val1) != len(val2):
            return False
        result = 0
        if isinstance(val1, bytes) and isinstance(val2, bytes):
            for x, y in zip(val1, val2):
                result |= x ^ y
        else:
            for x, y in zip(val1, val2):
                result |= ord(cast(str, x)) ^ ord(cast(str, y))
        return result == 0



from datetime import datetime


class ProductDetailsManager:
    def __init__(
        self,
        product_id: int = None,
        product_name: str = None,
        short_code: str = None,
        allow_trial: bool = None,
        trial_days: int = None,
        authorization_method: str = None,
        allow_overages: bool = None,
        max_overages: int = None,
        prevent_vm: bool = None,
        floating_timeout: int = None,
        metadata: dict = None,
    ) -> None:
        self.product_id = product_id
        self.product_name = product_name
        self.short_code = short_code
        self.allow_trial = allow_trial
        self.trial_days = trial_days
        self.authorization_method = authorization_method
        self.allow_overages = allow_overages
        self.max_overages = max_overages
        self.prevent_vm = prevent_vm
        self.floating_timeout = floating_timeout
        self.metadata = metadata

    def json_to_attribute(self, product_details: dict) -> None:
        for key, value in product_details.items():
            if key in self.__dict__:
                setattr(self, key, value)



from __future__ import annotations

from enum import Enum, auto


class ErrorType(Enum):
    """Licensefile error"""

    NO_LICENSEFILE = auto()
    CORRUPTED_LICENSEFILE = auto()

    """License status error"""

    LICENSE_NOT_ENABLED = auto()
    LICENSE_NOT_ACTIVE = auto()
    LICENSE_EXPIRED = auto()

    WRONG_LICENSE_TYPE = auto()
    WRONG_FEATURE_TYPE = auto()

    PRODUCT_MISMATCH = auto()
    HARDWARE_ID_MISMATCH = auto()

    VM_NOT_ALLOWED = auto()

    FLOATING_TIMEOUT = auto()

    CLOCK_TAMPERED = auto()

    NOT_ENOUGH_LICENSE_CONSUMPTIONS = auto()
    NOT_ENOUGH_FEATURE_CONSUMPTIONS = auto()
    NEGATIVE_CONSUMPTION_NOT_ALLOWED = auto()

    UNSUPPORTED_PRODUCT_FEATURE = auto()

    """activation"""

    OFFLINE_ACTIVATION_ERROR = auto()
    AIR_GAP_ACTIVATION_ERROR = auto()

    """Watchdogs"""

    LICENSE_WATCHDOG_ERROR = auto()
    FEATURE_WATCHDOG_ERROR = auto()


class LicenseSpringException(Exception):
    """Base exception class for all LicenseSpring-related errors."""

    def __init__(self, error_type: ErrorType, message):
        self.error_type = error_type
        self.message = message

    def __repr__(self):
        return "{}(error_name={},error_value={}, message={})".format(
            self.__class__.__name__,
            self.error_type.name,
            self.error_type.value,
            self.message,
        )

    def __str__(self) -> str:
        return f"{self.error_type} ({self.error_type.value}): {self.message}"


class LicenseActivationException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class LicenseDeleted(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class LicenseStateException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class LicenseSpringTypeError(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class ConfigurationMismatch(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class VMIsNotAllowedException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class TimeoutExpiredException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class ClockTamperedException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class ConsumptionError(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class ItemNotFoundError(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class WatchdogException(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)


class LicenseFileCorruption(LicenseSpringException):
    def __init__(self, error_type: ErrorType, message):
        super().__init__(error_type, message)



from __future__ import annotations

from datetime import datetime, timedelta, timezone

from licensespring.licensefile.error import (
    ConsumptionError,
    ErrorType,
    ItemNotFoundError,
    LicenseSpringTypeError,
)


class Feature:
    def __init__(self) -> None:
        self.floating_end_date = datetime.now(timezone.utc).replace(tzinfo=None)
        self.floating_start_date = datetime(year=2099, month=1, day=1)

        """
        Attributes:
            Feature fields: id,feature_type,expiry_date ...


        Methods:
            json_to_attribute(feature: dict): Converts JSON data into attributes for the Feature object
        """

    def json_to_attribute(self, feature: dict):
        """
        Converts JSON data into attributes for the Feature object. Special handling is applied to date fields
        to ensure they are parsed correctly from ISO format. If the feature is of type 'consumption' and doesn't
        already have a 'local_consumption' attribute, it initializes this attribute with 0.

        Args:
            feature (dict): A dictionary containing feature information, including any potential date fields and
            other attributes like 'feature_type' and 'local_consumption'.
        """
        for key, value in feature.items():
            if (
                key in ["expiry_date", "floating_end_date", "floating_start_date"]
                and value != None
            ):
                if value.endswith("Z"):
                    value = value[:-1]
                setattr(self, key, datetime.fromisoformat(value).replace(tzinfo=None))

            else:
                setattr(self, key, value)

        if self.feature_type == "consumption" and not hasattr(
            self, "local_consumption"
        ):
            self.local_consumption = 0

    def floating_end_datetime(self):
        if not hasattr(self, "floating_end_date"):
            setattr(
                self,
                "floating_end_date",
                datetime.now(timezone.utc).replace(tzinfo=None),
            )
        return self.floating_end_date

    def is_online_floating(self):
        return self.is_floating_cloud

    def is_offline_floating(self):
        return self.is_floating

    def floating_is_expired(self):
        if self.expiry_date == None or not hasattr(self, "expiry_date"):
            return False

        return datetime.now(timezone.utc).replace(tzinfo=None) > self.expiry_date


class FeatureManager:
    def __init__(self) -> None:
        """
        Initializes a new FeatureManager object that will hold and manage multiple Feature objects.

        Attributes:
            Feature objects

        Methods:
            return_features_list(): Returns lists of features
            remove_features(): Removes features that are not present in the latest feature response
            json_to_attribute(): JSON to attributes
            attributes_to_list(): Attributes to list
            add_local_consumption(feature_code, consumptions=1): Adds a specified number of consumptions to a feature
            register_feature(feature_code): Registers a feature fields as being in use.
            release_feature(self,feature_code): Releases a previously registered feature.

        """
        pass

    def return_features_list(self) -> list:
        return [key for key, feat in self.__dict__.items()]

    def remove_features(self, feature_response) -> None:
        """
        Removes features from the manager that are not present in the latest feature response.

        Args:
            feature_response (iterable): An iterable of dictionaries, each containing a 'code' key
            representing the feature code.

        """
        existing_features = set(self.return_features_list())

        new_features = set([feature["code"] for feature in feature_response])

        remove_features = existing_features - new_features

        for feature in remove_features:
            delattr(self, feature)

    def json_to_attribute(self, feature_response: dict):
        self.remove_features(feature_response)

        for feature in feature_response:
            if not hasattr(self, feature["code"]):
                setattr(self, feature["code"], Feature())

            feature_obj = getattr(self, feature["code"])

            feature_obj.json_to_attribute(feature)

    def attributes_to_list(self) -> list:
        feature_list = []

        for _, feat in self.__dict__.items():
            data = {}

            for key, value in feat.__dict__.items():
                if isinstance(value, datetime):
                    data[key] = value.isoformat()

                else:
                    data[key] = value

            feature_list.append(data)

        return feature_list

    def add_local_consumption(self, feature_code, consumptions=1) -> None:
        """
        Adds a specified number of local consumptions to a  feature, updating its consumption count.

        This method identifies a feature by its code and increments its local consumption by a given number. It performs several checks to validate the operation:
        1. It verifies the existence of the feature object. If the feature does not exist, it raises an `ItemNotFoundError`.
        2. It checks if the feature's type is set to "consumption". If not, a `LicenseSpringTypeError` is raised to indicate the feature is not of the consumption type.
        3. It determines whether the feature allows unlimited consumptions. If so, it directly increments the local consumption count.
        4. For limited consumption features, it checks if adding the specified consumptions would exceed the feature's maximum allowed consumptions and overages. If this limit would be exceeded, it raises a `ConsumptionError`.

        Args:
            feature_code (str): The code identifying the feature to which consumptions are to be added.
            consumptions (int, optional): The number of consumptions to add. Defaults to 1.

        Raises:
            ItemNotFoundError: If the feature specified by `feature_code` does not exist.
            LicenseSpringTypeError: If the identified feature is not of the "consumption" type.
            ConsumptionError: If adding the specified number of consumptions would exceed the feature's consumption limits.

        This method updates the `local_consumption` attribute of the specified feature object, reflecting the addition of the specified number of consumptions.
        """

        feature_obj = getattr(self, feature_code, None)

        if feature_obj == None:
            raise ItemNotFoundError(
                ErrorType.UNSUPPORTED_PRODUCT_FEATURE, "Features do not exists"
            )

        if feature_obj.feature_type != "consumption":
            raise LicenseSpringTypeError(
                ErrorType.WRONG_FEATURE_TYPE, "Feature not consumption type"
            )

        elif (
            getattr(feature_obj, "allow_negative_consumptions", True) == False
            and consumptions < 0
        ):
            raise ConsumptionError(
                error_type=ErrorType.NEGATIVE_CONSUMPTION_NOT_ALLOWED,
                message="Negative consumptions not allowed",
            )

        elif feature_obj.allow_unlimited_consumptions:
            feature_obj.local_consumption += consumptions

        elif (
            feature_obj.max_consumption + feature_obj.max_overages
            <= consumptions
            + feature_obj.total_consumptions
            + feature_obj.local_consumption
        ):
            raise ConsumptionError(
                ErrorType.NOT_ENOUGH_FEATURE_CONSUMPTIONS, "Not enough consumptions"
            )

        else:
            feature_obj.local_consumption += consumptions

    def register_feature(self, feature_code):
        """
        Registers a feature fields as being in use.

        Args:
            feature_code (str): The code of the feature to register as in use.
        """
        feature_obj = getattr(self, feature_code, None)

        if feature_obj.is_floating_cloud or feature_obj.is_floating:
            feature_obj.floating_start_date = datetime.now(timezone.utc).replace(
                tzinfo=None
            )
            feature_obj.floating_end_date = feature_obj.floating_start_date + timedelta(
                minutes=feature_obj.floating_timeout
            )

    def release_feature(self, feature_code: str):
        """
        Releases a previously registered feature.

        Args:
            feature_code (str): The code of the feature to release.
        """

        feature_obj = getattr(self, feature_code, None)

        if feature_obj.is_floating_cloud or feature_obj.is_floating:
            feature_obj.floating_start_date = datetime(year=2099, month=1, day=1)
            feature_obj.floating_end_date = datetime.now(timezone.utc).replace(
                tzinfo=None
            )
            if feature_obj.floating_in_use_devices > 0:
                feature_obj.floating_in_use_devices -= 1

    def get_feature_dict(self, feature_code: str) -> dict:
        """
        Get feature data inside dictionary

        Args:
            feature_code (str): feature short code

        Returns:
            dict: feature data
        """

        feature_obj = getattr(self, feature_code, None)

        if feature_obj:
            return feature_obj.__dict__



import requests

from licensespring.licensefile import License
from licensespring.licensefile.base_manager import BaseManager
from licensespring.licensefile.config import Configuration


class FloatingManager(BaseManager):
    def __init__(self, conf: Configuration) -> None:
        super().__init__(conf=conf)

    def auth(self, username: str, password: str) -> dict:
        """
        Authenitcate

        Args:
            username (str): username
            password (str): password

        Returns:
            dict: response
        """
        return self.api_client.floating_client.auth(
            username=username, password=password
        )

    def register(
        self,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
        license_id: int = None,
    ) -> License:
        """
        Register License

        Args:
            os_hostname (str, optional): os hostname. Defaults to None.
            ip_local (str, optional): ip local. Defaults to None.
            user_info (str, optional): user info. Defaults to None.
            license_id (int, optional):license id. Defaults to None.

        Returns:
            License: license
        """

        response = self.api_client.floating_client.register_user(
            product=self._conf._product,
            license_id=license_id,
            os_hostname=os_hostname,
            ip_local=ip_local,
            user_info=user_info,
        )
        self.licensefile_handler._cache.update_cache(
            "floating_server_register", response
        )
        self.licensefile_handler._cache.update_floating_period(
            self.licensefile_handler._cache.borrowed_until
        )
        self.licensefile_handler.save_licensefile()

        return License(
            self._conf._product,
            api_client=self.api_client,
            licensefile_handler=self.licensefile_handler,
        )

    def unregister(self, license_id: int = None) -> str:
        """
        Unregister license

        Args:
            license_id (int, optional): license id. Defaults to None.

        Returns:
            str: license unregistered
        """
        return self.api_client.floating_client.unregister_user(
            product=self._conf._product, license_id=license_id
        )

    def unregister_all(self) -> str:
        """
        Unregister all

        Returns:
            str: ""
        """
        return self.api_client.floating_client.unregister_all()

    def borrow(
        self,
        borrowed_until: str,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
        license_id: int = None,
    ) -> License:
        """
        borrow license

        Args:
            borrowed_until (str): borrow until date
            os_hostname (str, optional): os hostname. Defaults to None.
            ip_local (str, optional): ip local. Defaults to None.
            user_info (str, optional): user info. Defaults to None.
            license_id (int, optional):license id. Defaults to None.

        Returns:
            License: license
        """
        response = self.api_client.floating_client.register_user(
            product=self._conf._product,
            license_id=license_id,
            os_hostname=os_hostname,
            ip_local=ip_local,
            user_info=user_info,
        )
        self.licensefile_handler._cache.update_cache(
            "floating_server_register", response
        )
        self.licensefile_handler.save_licensefile()

        license = License(
            self._conf._product,
            api_client=self.api_client,
            licensefile_handler=self.licensefile_handler,
        )

        license.floating_borrow(borrow_until=borrowed_until)

        return license

    def is_online(self, throw_e=False) -> bool:
        """
        Checks if floating server is online

        Args:
            throw_e (bool, optional): True if you want raise exception. Defaults to False.

        Raises:
            ex: Exception

        Returns:
            bool: True if server is online, otherwise False
        """
        try:
            response = requests.get(
                f"{self.api_client.api_protocol}://{self.api_client.api_domain}"
            )

            if response.status_code == 200:
                return True

        except Exception as ex:
            if throw_e:
                raise ex

        return False



from __future__ import annotations

from licensespring.api.signature import SignatureVerifier
from licensespring.hardware import HardwareIdProvider


class Configuration:
    """
    A class to configure settings for a license management system.

    This class encapsulates various configuration settings used for managing
    licenses, including authentication keys, encryption settings, hardware ID
    providers, signature verification, and API details.

    Attributes:
        product (str): Name of the product.
        api_key (str): API key for authentication.
        shared_key (str): Shared key for additional security.
        file_key (str): Encryption key for license files. Default is a sample key.
        file_iv (str): Initialization vector for encryption. Default is a sample vector.
        hardware_id_provider (object): Provider class for hardware ID. Default is HardwareIdProvider.
        verify_license_signature (bool): Flag to enable/disable signature verification. Default is True.
        signature_verifier (object): Verifier class for checking signatures. Default is SignatureVerifier.
        api_domain (str): Domain for the API server. Default is "api.licensespring.com".
        api_version (str): Version of the API. Default is "v4".
        filename (str): Name for the license file. Default is "License".
        file_path (str): Path to save the license file. Default is None.
        grace_period_conf (int): Grace period configuration in days. Default is 12.
        is_guard_file_enabled (bool): Enables guard protection for offline licenses if set to True.
    """

    def __init__(
        self,
        product: str,
        api_key: str,
        shared_key: str,
        file_key: str,
        file_iv: str,
        hardware_id_provider=HardwareIdProvider,
        verify_license_signature=True,
        signature_verifier=SignatureVerifier,
        api_domain="api.licensespring.com",
        api_protocol="https",
        api_version="v4",
        filename="License",
        file_path=None,
        grace_period_conf=24,
        is_guard_file_enabled=True,
        air_gap_public_key=None,
    ) -> None:
        self._product = product

        self._api_key = api_key
        self._shared_key = shared_key

        self._file_key = file_key
        self._file_iv = file_iv

        self._hardware_id_provider = hardware_id_provider

        self._verify_license_signature = verify_license_signature
        self._signature_verifier = signature_verifier
        self._api_domain = api_domain
        self._api_version = api_version
        self._api_protocol = api_protocol

        self._filename = filename
        self._file_path = file_path
        self.grace_period_conf = grace_period_conf
        self.is_guard_file_enabled = is_guard_file_enabled
        self.air_gap_public_key = air_gap_public_key



import logging
from base64 import b64decode, b64encode
from datetime import datetime, timezone

from Crypto.Hash import SHA1, SHA256
from Crypto.PublicKey import ECC
from Crypto.Signature import DSS
from Crypto.Util.number import bytes_to_long, long_to_bytes

from licensespring.api import APIClient
from licensespring.floating_server import FloatingAPIClient
from licensespring.licensefile.config import Configuration
from licensespring.licensefile.error import ErrorType, LicenseActivationException
from licensespring.licensefile.offline_activation_guard import (
    OfflineActivation,
    OfflineActivationGuard,
)


class APIService(APIClient):
    def __init__(self, conf: Configuration) -> None:
        super().__init__(
            api_key=conf._api_key,
            shared_key=conf._shared_key,
            hardware_id_provider=conf._hardware_id_provider,
            verify_license_signature=conf._verify_license_signature,
            signature_verifier=conf._signature_verifier,
            api_domain=conf._api_domain,
            api_version=conf._api_version,
            api_protocol=conf._api_protocol,
        )

        self._conf = conf
        self.floating_client = FloatingAPIClient(
            api_protocol=conf._api_protocol,
            api_domain=conf._api_domain,
            api_version=conf._api_version,
        )

    def create_activation_guard(self, id: str) -> OfflineActivationGuard:
        """
        Creates OfflineActivationGuard object

        Args:
            id (str): response_id

        Returns:
            OfflineActivationGuard
        """

        offline_guard = OfflineActivationGuard()
        offline_guard.set_id(id=id)
        offline_guard.set_device_id(self.hardware_id_provider.get_id())
        offline_guard.set_date_created()

        return offline_guard

    def check_guard(self, offline_data: OfflineActivation, id: str):
        """
        Checks the validity of the .lic file

        Args:
            offline_data (OfflineActivation): OfflineActivation object

        Raises:
            LicenseActivationException: Activation data is not valid
            LicenseActivationException: Response file ID mismatch
            LicenseActivationException: License does not belong to this device
        """

        if offline_data._guard._date_created > datetime.now(timezone.utc).replace(
            tzinfo=None
        ):
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "Activation data is not valid, please restart activation process.",
            )

        elif offline_data._guard._id != id:
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "Response file ID mismatch, please restart activation process.",
            )

        elif offline_data._guard._device_id != self._conf._hardware_id_provider.get_id(
            self
        ):
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "License does not belong to this device.",
            )

    def get_air_gap_activation_code(
        self, initialization_code: str, license_key: str
    ) -> str:
        """
        Get air gap activation code

        Args:
            initialization_code (str): initialization code
            license_key (str): license key

        Returns:
            str: activation code
        """
        hardware_id = self.hardware_id_provider.get_id()
        message = (hardware_id + license_key).encode("utf-8")
        decoded_key = b64decode(initialization_code)

        private_key = ECC.construct(curve="P-256", d=bytes_to_long(decoded_key))

        digest = SHA1.new(message)

        signer = DSS.new(private_key, "deterministic-rfc6979")
        signature = signer.sign(digest)

        r, s = int.from_bytes(
            signature[: len(signature) // 2], byteorder="big"
        ), int.from_bytes(signature[len(signature) // 2 :], byteorder="big")

        degree = private_key.pointQ.size_in_bits()
        bn_len = (degree + 7) // 8
        r_bytes = long_to_bytes(r, bn_len)
        s_bytes = long_to_bytes(s, bn_len)

        raw_buf = r_bytes + s_bytes

        return b64encode(raw_buf).decode("utf-8")

    def verify_confirmation_code(
        self, confirmation_code: str, license_key: str, policy_id: str
    ) -> bool:
        """
        Verify confirmation code

        Args:
            confirmation_code (str): confirmation code
            license_key (str): license key
            policy_id (str): policy id

        Returns:
            bool: True if verification was successful, otherwise False.
        """
        hardware_id = self.hardware_id_provider.get_id()
        message = (hardware_id + license_key + policy_id).encode("utf-8")
        decoded_key = b64decode(self._conf.air_gap_public_key)

        hex_key = decoded_key.hex()
        gx = int(hex_key[:64], 16)
        gy = int(hex_key[64:], 16)

        public_key = ECC.construct(curve="P-256", point_x=gx, point_y=gy)
        decoded_signature = b64decode(confirmation_code)

        signature = DSS.new(public_key, "fips-186-3")
        digest = SHA256.new(message)

        try:
            signature.verify(digest, decoded_signature)
            return True
        except (ValueError, TypeError):
            logging.info("Failed to verify EC Signature")
            return False

    def activate_air_gapped_licenses(
        self, data: OfflineActivation, license_key: str, policy_id: str
    ) -> dict:
        """
        Activate air gapped license

        Args:
            data (OfflineActivation): Offline activation data
            license_key (str): license key
            policy_id (str): policy id

        Raises:
            LicenseActivationException: Policy ID mismatch

        Returns:
            dict: response
        """
        response = data.decode_offline_activation()
        logging.info("Offline file activation id:", response["id"])

        if str(response["id"]) != policy_id:
            raise LicenseActivationException(
                ErrorType.AIR_GAP_ACTIVATION_ERROR,
                "Policy file ID mismatch, please use another file.",
            )

        if data._use_guard:
            logging.info("Checking activation guard")
            self.check_guard(data, id=license_key)
            logging.info("Activation guard check passed")

        response["license_key"] = license_key
        response["license_type"] = response["default_license_type"]
        response["policy_id"] = response["id"]

        del response["default_license_type"]
        # Next line insures that product is not overwritten with product_id
        del response["product"]

        return response

    def activate_license_offline(self, data: OfflineActivation) -> dict:
        """
        Decodes the data and checks the guard file

        Args:
            data (OfflineActivation): offline activation data

        Returns:
            dict: response
        """
        response = data.decode_offline_activation()

        logging.info("Offline file activation id:", response["request_id"])

        if data._use_guard:
            logging.info("Checking activation guard")
            self.check_guard(data, response["request_id"])
            logging.info("Activation guard check passed")

        return response



from licensespring.licensefile import License
from licensespring.licensefile.config import Configuration
from licensespring.licensefile.data_storage import DataStorage
from licensespring.licensefile.error import LicenseFileCorruption
from licensespring.licensefile.service import APIService


class BaseManager:
    def __init__(self, conf: Configuration):
        self._conf = conf
        self.licensefile_handler = DataStorage(conf)
        self.api_client = APIService(conf=conf)

    def load_license(self) -> License:
        """
        Loads licensefile and sets attributes for LicenseData instance
        Returns:
        License: An instance of the License class reflecting the loaded license.
        """
        self.licensefile_handler.load_licensefile()

        return License(self._conf._product, self.api_client, self.licensefile_handler)

    def current_config(self) -> dict:
        """
        Get current configuration

        Returns:
            dict: configuration
        """
        return self._conf.__dict__

    def reconfigure(self, conf: Configuration) -> None:
        """
        Reconfigure

        Args:
            conf (Configuration): Configuration
        """
        self._conf = conf
        self.licensefile_handler = DataStorage(conf)
        self.api_client = APIService(conf=conf)

    def is_license_file_corrupted(self) -> bool:
        """
        Check if licensefile is corrupted

        Returns:
            bool: True if licensefile is corrupted otherwise False
        """
        try:
            self.licensefile_handler.load_licensefile()
            return False
        except LicenseFileCorruption:
            return True

    def clear_local_storage(self):
        """
        Clear all data from current product
        """
        self.licensefile_handler.clear_storage()

    def data_location(self) -> str:
        """
        Get licensefile location

        Returns:
            str: licensefile location
        """

        return self.licensefile_handler.license_path

    def set_data_location(self, path: str):
        """
        Set data location

        Args:
            path (str): new data location path
        """
        self.licensefile_handler._license_path = path

    def license_file_name(self) -> str:
        """
        Get licensefile name

        Returns:
            str: licensefile name
        """

        return self.licensefile_handler._filename

    def set_license_file_name(self, name: str):
        """
        Set licensefile name

        Args:
            name (str): license file name
        """

        self.licensefile_handler._filename = name



class Customer:
    def __init__(
        self,
        email,
        first_name="",
        last_name="",
        company="",
        phone="",
        address="",
        city="",
        state="",
        country="",
        postcode="",
        reference="",
        metadata={},
    ):
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.company = company
        self.phone = phone
        self.address = address
        self.city = city
        self.state = state
        self.country = country
        self.postcode = postcode
        self.reference = reference
        self.metadata = metadata



import json

import requests
from requests.auth import HTTPBasicAuth

import licensespring
from licensespring.api.error import ClientError
from licensespring.hardware import HardwareIdProvider


class FloatingAPIClient:
    def __init__(
        self,
        api_protocol: str = "http",
        api_domain: str = "localhost",
        api_version: str = "v4",
        hardware_id_provider: HardwareIdProvider = HardwareIdProvider,
    ):
        self.bearer_token = None
        self.hardware_id_provider = hardware_id_provider()
        self.api_protocol = api_protocol
        self.api_domain = api_domain
        self.api_version = api_version
        self.api_base = f"{api_protocol}://{api_domain}/api/{api_version}"

    def api_url(self, endpoint: str) -> str:
        return f"{self.api_base}{endpoint}"

    def auth(
        self,
        username: str,
        password: str,
    ) -> dict:
        """
        Authenticate

        Args:
            username (str): username
            password (str): password

        Returns:
            dict: response
        """
        data = {"username": username, "password": password}

        response = requests.request(
            method="get",
            url=f"{self.api_protocol}://{self.api_domain}/auth",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            auth=HTTPBasicAuth(username, password),
        )

        if 400 <= response.status_code < 500:
            raise ClientError(response)
        else:
            response.raise_for_status()

        data = response.json()
        self.bearer_token = data.get("token")

        return data

    def request_hardware_data(
        self,
        data: dict,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
    ) -> dict:
        """
        Request hardware data

        Args:
            data (dict): data
            os_hostname (str, optional): os hostname. Defaults to None.
            ip_local (str, optional): ip local. Defaults to None.
            user_info (str, optional): user info. Defaults to None.

        Returns:
            dict: data
        """
        data["ip_local"] = ip_local
        data["user_info"] = user_info

        data["os_hostname"] = (
            os_hostname
            if os_hostname is not None
            else self.hardware_id_provider.get_hostname()
        )

        return data

    def request_version_data(self, data: dict, app_ver: str = None) -> dict:
        """
        Version data

        Args:
            data (dict): data
            app_ver (str, optional): app version. Defaults to None.

        Returns:
            dict: response
        """
        data["sdk_ver"] = f"PythonSDK {licensespring.version}"

        if app_ver:
            data["app_ver"] = app_ver
        elif licensespring.app_version:
            data["app_ver"] = licensespring.app_version

        return data

    def request_floating_data(
        self, data: dict, product: str, license_id: int = None, user: str = None
    ) -> dict:
        """
        request data for floating server communication

        Args:
            data (dict): data
            product (str): product short code
            license_id (int, optional): license id. Defaults to None.
            user (str, optional): hardware id. Defaults to None.

        Returns:
            dict: response
        """
        data["product"] = product
        data["user"] = user if user else self.hardware_id_provider.get_id()

        if license_id:
            data["license_id"] = license_id

        return data

    def request_additional_data(self, data: dict, additional_data: dict) -> dict:
        """
        request additional data
        Args:
            data (dict): data
            additional_data (dict): additional data

        Returns:
            dict: data
        """
        for key, value in additional_data.items():
            if value is not None:
                data[key] = value
        return data

    def request_headers(self, custom_headers: dict = {}) -> dict:
        """
        request headers

        Args:
            custom_headers (dict, optional): custom headers. Defaults to {}.

        Returns:
            dict: headers
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}

        if self.bearer_token:
            headers["Authorization"] = f"Bearer {self.bearer_token}"
        return {**headers, **custom_headers}

    def send_request(
        self,
        method: str,
        endpoint: str,
        custom_headers: dict = {},
        params: dict = None,
        data: dict = None,
        json_data: json = None,
    ):
        """
        Send request to FS server

        Args:
            method (str): method
            endpoint (str): endpoint
            custom_headers (dict, optional): custom headers. Defaults to {}.
            params (dict, optional): params. Defaults to None.
            data (dict, optional): data. Defaults to None.
            json_data (json, optional): json data. Defaults to None.

        Raises:
            ClientError: Error from FS

        Returns:
            dict: response
        """
        response = requests.request(
            method=method,
            url=self.api_url(endpoint),
            headers=self.request_headers(custom_headers=custom_headers),
            params=params,
            data=data,
            json=json_data,
        )
        if 400 <= response.status_code < 500:
            raise ClientError(response)
        else:
            response.raise_for_status()
        return response

    def register_user(
        self,
        product: str,
        user: str = None,
        license_id: int = None,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
    ) -> dict:
        """
        Register license

        Args:
            product (str): product short code
            user (str): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.
            os_hostname (str, optional): hostname. Defaults to None.
            ip_local (str, optional): ip local. Defaults to None.
            user_info (str, optional): user info. Defaults to None.

        Returns:
            dict: response
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data = self.request_hardware_data(
            data=data, os_hostname=os_hostname, ip_local=ip_local, user_info=user_info
        )

        response = self.send_request(
            method="post", endpoint="/register", json_data=data
        )

        # Currently there is no signature verification

        return response.json()

    def unregister_user(
        self, product: str, user: str = None, license_id: int = None
    ) -> str:
        """
        unregister license

        Args:
            product (str): product short code
            user (str): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.

        Returns:
            str: "user_unregistered"
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        response = self.send_request(
            method="post", endpoint="/unregister", json_data=data
        )

        return response.content.decode()

    def unregister_all(self) -> str:
        """
        Unregister all

        Returns:
            str: ""
        """
        response = self.send_request(method="get", endpoint="/unregister/all")

        return response.content.decode()

    def borrow(
        self,
        product: str,
        borrowed_until: str,
        user: str = None,
        license_id: int = None,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
    ) -> dict:
        """
        Borrow license

        Args:
            product (str): product short code
            borrowed_until (str): borrow until (e.g. 2029-05-06T00:00:00Z)
            user (str, optional): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.
            os_hostname (str, optional): os hostname. Defaults to None.
            ip_local (str, optional): local ip. Defaults to None.
            user_info (str, optional): user info. Defaults to None.

        Returns:
            dict: response
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data["borrowed_until"] = borrowed_until
        data = self.request_hardware_data(
            data=data, os_hostname=os_hostname, ip_local=ip_local, user_info=user_info
        )

        response = self.send_request(method="post", endpoint="/borrow", json_data=data)

        return response.json()

    def add_consumption(
        self,
        product: str,
        consumptions: int = 1,
        max_overages: int = None,
        allow_overages: bool = None,
        user: str = None,
        license_id: int = None,
    ) -> dict:
        """
        Add license consumptions

        Args:
            product (str): product short code
            consumptions (int, optional): consumptions. Defaults to 1.
            max_overages (int, optional): max overages. Defaults to None.
            allow_overages (bool, optional): allow overages. Defaults to None.
            user (str, optional): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.

        Returns:
            dict: response
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data = self.request_additional_data(
            data=data,
            additional_data={
                "consumptions": consumptions,
                "max_overages": max_overages,
                "allow_overages": allow_overages,
            },
        )

        response = self.send_request(
            method="post", endpoint="/add_consumption", json_data=data
        )

        return response.json()

    def add_feature_consumption(
        self,
        product: str,
        feature_code: str,
        consumptions: int = 1,
        user: str = None,
        license_id: int = None,
    ) -> dict:
        """
        Add feature consumption

        Args:
            product (str): product short code
            feature_code (str): feature code
            user (str, optional): user (default uses hardware id)
            consumptions (int, optional): consumptions. Defaults to 1.
            license_id (int, optional): license id. Defaults to None.

        Returns:
            dict: response
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data = self.request_additional_data(
            data=data,
            additional_data={
                "consumptions": consumptions,
                "feature": feature_code,
            },
        )

        response = self.send_request(
            method="post", endpoint="/add_feature_consumption", json_data=data
        )

        return response.json()

    def feature_register(
        self,
        product: str,
        feature_code: str,
        user: str = None,
        license_id: int = None,
        borrow_until: str = None,
        os_hostname: str = None,
        ip_local: str = None,
        user_info: str = None,
    ) -> dict:
        """
        Feature register

        Args:
            product (str): product short code
            feature_code (str): feature short code
            user (str, optional): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.
            borrowed_until (str): borrow until (e.g. 2029-05-06T00:00:00Z)
            os_hostname (str, optional): os hostname. Defaults to None.
            ip_local (str, optional): ip local. Defaults to None.
            user_info (str, optional): user info. Defaults to None.

        Returns:
            dict: response
        """

        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data = self.request_additional_data(
            data=data,
            additional_data={"feature": feature_code, "borrow_until": borrow_until},
        )
        data = self.request_hardware_data(
            data=data, os_hostname=os_hostname, ip_local=ip_local, user_info=user_info
        )

        response = self.send_request(
            method="post", endpoint="/featureRegister", json_data=data
        )

        return response.json()

    def feature_release(
        self, product: str, feature_code: str, user: str = None, license_id: int = None
    ) -> str:
        """
        Feature release

        Args:
            product (str): product short code
            feature_code (str): feature short code
            user (str, optional): user (default uses hardware id)
            license_id (int, optional): license id. Defaults to None.

        Returns:
            str: "feature_released"
        """
        data = self.request_floating_data(
            data={},
            product=product,
            user=user,
            license_id=license_id,
        )
        data = self.request_version_data(data=data)
        data = self.request_additional_data(
            data=data,
            additional_data={"feature": feature_code},
        )

        response = self.send_request(
            method="post", endpoint="/featureRelease", json_data=data
        )

        return response.content.decode()



from datetime import datetime


class VariablesManager:
    def __init__(self) -> None:
        pass

    def json_to_attribute(self, device_vars: list) -> None:
        for var in device_vars:
            if var["created_at"].endswith("Z"):
                var["created_at"] = var["created_at"][:-1]

            setattr(
                self,
                var["variable"],
                DeviceVariable(
                    id=var.get("id", None),
                    created_at=datetime.fromisoformat(var["created_at"]).replace(
                        tzinfo=None
                    ),
                    variable=var.get("variable", None),
                    value=var.get("value", None),
                    device_id=var.get("device_id", None),
                    update=False,
                ),
            )

    def set_variables(self, variables: dict):
        """
        Sets device variables in cache

        Args:
            variables (dict): device variables
        """
        for variable in variables.keys():
            setattr(
                self,
                variable,
                DeviceVariable(
                    id=None,
                    created_at=None,
                    variable=variable,
                    value=variables[variable],
                    device_id=None,
                    update=True,
                ),
            )

    def attributes_to_list(self) -> list:
        """
        Generates list of device variables

        Returns:
            list: device variables
        """
        device_varaibles_list = []

        for _, device_var in self.__dict__.items():
            data = {}

            for key, value in device_var.__dict__.items():
                if key != "update":
                    if isinstance(value, datetime):
                        data[key] = value.isoformat()

                    else:
                        data[key] = value

            device_varaibles_list.append(data)

        return device_varaibles_list

    def get_device_variable(self, variable_name: str) -> dict:
        """
        Get device variable

        Args:
            variable_name (str): device variable name

        Returns:
            dict: device variable
        """
        if hasattr(self, variable_name):
            obj = getattr(self, variable_name)

            return obj.__dict__

    def get_device_variable_for_send(self) -> dict:
        """
        Finds new device variables which should be sent on LicenseSpring server

        Returns:
            dict: Dictionary of device variables
        """
        new_device_variables = {}

        for _, device_var in self.__dict__.items():
            if device_var.update:
                new_device_variables[device_var.variable] = device_var.value
                device_var.update = False

        return new_device_variables


class DeviceVariable:
    def __init__(
        self,
        id: int,
        created_at: str,
        variable: str,
        value: str,
        device_id: int,
        update: bool,
    ) -> None:
        self.id = id
        self.created_at = created_at
        self.variable = variable
        self.value = value
        self.device_id = device_id
        self.update = update



import base64
import json
from datetime import datetime, timezone


class OfflineActivationGuard:
    def __init__(self) -> None:
        pass

    def set_id(self, id: str):
        self._id = id

    def set_device_id(self, device_id: str):
        self._device_id = device_id

    def set_date_created(self):
        self._date_created = datetime.now(timezone.utc).replace(tzinfo=None)

    def to_json(self) -> str:
        data = {}
        for key, value in self.__dict__.items():
            if isinstance(value, datetime):
                data[key] = value.isoformat()

            else:
                data[key] = value

        return json.dumps(data)

    @classmethod
    def from_json(cls, data: dict):
        obj = cls()

        for key, value in data.items():
            if key == "_date_created":
                if value.endswith("Z"):
                    value = value[:-1]

                setattr(obj, key, datetime.fromisoformat(value).replace(tzinfo=None))

            else:
                setattr(obj, key, value)
        return obj


class OfflineActivation:
    def __init__(self) -> None:
        self._guard = None

    def set_data(self, data: str):
        self._data = data

    def set_use_guard(self, use_guard: bool):
        self._use_guard = use_guard

    def set_is_activation(self, is_activation: bool):
        self._is_activation = is_activation

    def decode_offline_activation(self):
        decoded_data = base64.b64decode(self._data).decode()
        return json.loads(decoded_data)

    def create_guard_file(self, offline_activation: OfflineActivationGuard):
        self._guard = offline_activation
